//! Implementations of [`Arbitrary`] for [`core`] types.
//!
//! [`Arbitrary`]: crate::Arbitrary

mod array;
mod bool;
mod cell;
mod char;
mod cmp;
mod iter;
mod marker;
mod num;
mod ops;
mod option;
mod result;
mod slice;
mod str;
mod sync;
mod time;
mod tuple;
mod unit;
